
/*

Andreja Andric

One Bit Net Dance

For live performance of Concert for Computer Network and One Bit Motion Mix

Includes adaptations for live performance of Malgorzata Zurada's Feeding Celestial Bodies performance
and network interface for collective network performance of a laptop ensemble

April 5th 2015 - April 2020

To perform live, compile this program with java compiler and run it.

You need two config files, named input0.txt and ClientConfiguration.txt.

The input0.txt config file contains 5 integers, m fractions and n integers, where m, n>0. 
Their meanings are:

Main Pitch = Frequency of the main tone
Duration = durations of tones
Length = length of tones
Decay = Inverse Decay Rate
ScaleType = (6 = 6-harmonic scale with 7 notes per octave, 10 = 10-harmonic scale with 19 notes per octave)
PolyphonyType = 1/1 3/2 -2/1 = Array of fractions indicating fixed relations between individual voices. 
Positive numbers indicate parallel motion, negative numbers indicate contrary motion.
A shortcut can be used here for some kinds of two voice polyphony: 
0 for contrary motion (1/1 -1/1), 1 for parallel octaves (1/1 1/2) and 2 for parallel fifths (1/1 2/3)
The notes from different voices are played arpeggio with the distance indicated by Duration above.
Remaining n numbers form the melody.
Underscore instead of a number denotes a break.

Textual comments may be mixed with numbers. They are ignored by the program.

Example:

Center frequency [Hz]: 160
Beat [ms]: 64
Reverb [ms]: 200 
Inverse Decay Rate: 4
Scale type: 6
Polyphony type: 1/1 3/4 1/2

5 4 0 4 5 0 5 0 5 4 _ 5 4 _ 4 5


During performance, repeatedly change the numbers in config files and save them.

Stop with 0

The other config file contains three lines:
- IP Address of the server
- your nickname that will appear in a network performance interface
- whether your computer is muted or not; namely, the sound result of a group performance
is one, not many and only one of the computers should be amplified, whereas the others should
be muted; although all the computers maintain the joint melody.

*/

import java.util.Scanner;
import java.util.Arrays;
import java.util.Vector;
import java.util.LinkedList;
import java.nio.file.FileSystems;
import java.nio.file.Files;
import java.nio.file.Path;
import java.io.IOException;
import java.util.NoSuchElementException;
import java.io.*;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.SourceDataLine;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.Graphics;
import java.awt.Color;
import java.awt.Font;
import java.awt.event.KeyEvent;
import java.util.Timer;
import java.util.TimerTask;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JButton;
import javax.swing.JTextArea;
import javax.swing.DefaultListModel;
import javax.swing.JMenuBar;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.KeyStroke;
import javax.swing.SwingConstants;
import java.awt.event.WindowEvent;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.Toolkit;
import java.net.*;

public class OneBitNetDance extends JPanel implements Runnable  
{
    public static final int MaxPlayers = 10;

    //These three variables and the function below form the linear congruent random number generator that the program
    //uses to make micro-decisions while it generates the current clip.
    int a, b, c;

    int next() { a=(a*b%Integer.MAX_VALUE+c)%Integer.MAX_VALUE; if(a<0) a=-a; return a; }

    private class StringPair
    {
       public String name, content;

       public StringPair(String n, String ct)
       {
          name = n;
          content = ct;
       }
    }

    private class NetDanceListener implements Runnable
    {
        private boolean NewState;
        private int MainPitch, Duration, Length, DecayRate, ScaleType;
        private int[] PolyphonyType;
        private String PolyphonyTypeStr;
        private int[] CompoundMelody;
        private String CompoundMelodyString;
        private String[] NameList;
        private int[] IndexList;
        private LinkedList<StringPair> qReq;
        private boolean alive;
        private int[] Melody;

        public boolean isNewState() { return NewState; }
        public void UpdateState() { NewState = false; }
        public int getMainPitch() { return MainPitch; }
        public int getDuration()  { return Duration; }
        public int getLength()    { return Length; }
        public int getDecayRate() { return DecayRate; }
        public int getScaleType() { return ScaleType; }
        public int[] getPolyphonyType() { return PolyphonyType; }
        public int[] getCompoundMelody() { return CompoundMelody; } 
        public String getCompoundMelodyString() { return CompoundMelodyString; }
        public String[] getNameList() { return NameList; } 
        public int[] getIndexList() { return IndexList; } 
        public int getPerformerCount() { return NameList.length; }

        public NetDanceListener()
        {
           alive = true;
           NewState = true;
           Melody = new int[4];
           for(int j=0;j<4;j++)
              Melody[j] = 0;
           PolyphonyType = new int[] {1, 1, 1, 1, 2, 1};
      
           CompoundMelody = Melody;
           NameList = null;
           IndexList = null;
           qReq = new LinkedList<StringPair>();
        }

        public void kill()
        {
           alive = false;
        }

        private void ParsePolyphonyType(String PTString)
        {
           boolean err = false;
           int j=0;
           String[] fractions = PTString.split("\\s+");
           int[] newPolyphonyType = new int[3*fractions.length];
           for(int i=0;i<fractions.length&&!err;i++)
           {
              String[] components = fractions[i].split("/");
              try 
              {
                 newPolyphonyType[j++] = Math.abs(Integer.parseInt(components[0]));
                 newPolyphonyType[j++] = Integer.parseInt(components[1]);
                 newPolyphonyType[j++] = (fractions[i].charAt(0)=='-'?-1:1);
              }
              catch(NumberFormatException e)
              {
                 System.out.println("E001: Error in polyphony type definitions: "+fractions[i]);
                 err = true;
              }
           }
           if(!err)
              PolyphonyType = newPolyphonyType;
        }

        private void ParseCompoundMelody(String MelodyString)
        {
           boolean err = false;
           String[] notes = MelodyString.split("\\s+");
           int[] newCompoundMelody = new int[notes.length];
           for(int i=0;i<notes.length&&!err;i++)
           {
              try 
              {
                 newCompoundMelody[i] = Integer.parseInt(notes[i]);
              }
              catch(NumberFormatException e)
              {
                 System.out.println("E002: Error in melody definition, note " + i);
                 err = true;
              }
           }
           if(!err)
              CompoundMelody = newCompoundMelody;
           else
              CompoundMelody = Melody;
        }

        private void GetStateFromServer() throws IOException
        {
           int i;
           if(NetworkEstablished)
           {
              out.println("get");
              String fromServer = in.readLine();
              if(fromServer.equals("N"))
              {
                 //the window does not refresh the first time and ends up empty for no reason.
                 //repainting the last line makes sure it does repaint. Though I still don't understand.
                 int j=everyPlayerMelodyWindow.getLineCnt()-1;
                 everyPlayerMelodyWindow.setParticipantInput("-----------------", j);
                 return;
              }
              if(!fromServer.equals("more lines"))
              {
                 System.out.println("E003: Error in header in response from Server: "+fromServer);
                 return;
              }
              String CommonPar = "MainPitch";
              try
              {
                 fromServer = in.readLine();
                 MainPitch = Integer.parseInt(fromServer);
                 CommonPar = "Duration";
                 fromServer = in.readLine();
                 Duration = Integer.parseInt(fromServer);
                 CommonPar = "Length";
                 fromServer = in.readLine();
                 Length = Integer.parseInt(fromServer);
                 CommonPar = "DecayRate";
                 fromServer = in.readLine();
                 DecayRate = Integer.parseInt(fromServer);
                 CommonPar = "ScaleType";
                 fromServer = in.readLine();
                 ScaleType = Integer.parseInt(fromServer);
              }
              catch( NumberFormatException e)
              {
                 System.out.println("E004: Could not parse common parameter "+CommonPar+" in response from Server. Received: "+fromServer);
                 return;
              }
              fromServer = in.readLine();
              PolyphonyTypeStr = fromServer;
              ParsePolyphonyType(fromServer);
              fromServer = in.readLine();
              CompoundMelodyString = fromServer;
              ParseCompoundMelody(fromServer);

              int performerCount = 0;
              try
              {
                 fromServer = in.readLine();
                 performerCount = Integer.parseInt(fromServer);
              }
              catch( NumberFormatException e)
              {
                 System.out.println("E008: Could not parse the performers count in response from Server: "+fromServer);
                 return;
              }
              NameList = new String[performerCount];
              IndexList = new int[performerCount];
              StringBuilder sb = new StringBuilder();
              int j=0;
              for(i=0;i<performerCount;i++)
              {
                 fromServer="";//name
                 fromServer = in.readLine();
                 NameList[i] = fromServer;
                 everyPlayerMelodyWindow.setParticipantInput(fromServer, j++);
                 fromServer="";
                 fromServer = in.readLine();
                 int currentIndex;
                 try 
                 {
                    currentIndex = Integer.parseInt(fromServer);
                 }
                 catch(NumberFormatException e)
                 {
                    System.out.println("E002: Index missing, found " + fromServer);
                    return;
                 }
                 IndexList[i] = currentIndex;
                 while(!fromServer.equals("end part"))
                 {
                    fromServer = in.readLine();
                    if(!fromServer.equals("end part"))
                    {
                       if(currentIndex==Identifier&&j!=0) 
                          sb.append(System.lineSeparator());
                       //this may be happening during the initialization, when the windows have
                       //not been created and populated yet
                       if(fromServer.indexOf(":")==-1)
                          everyPlayerMelodyWindow.setParticipantInput(fromServer, j++);
                       if(currentIndex==Identifier)
                          sb.append(fromServer);
                    }
                    else
                    {
                       if(currentIndex==Identifier)
                          compoundMelodyWindow.updateCommandArea(sb.toString());
                    }
                 }
              }
              fromServer = in.readLine();
              if(!fromServer.equals("end lines"))
              {     
                 System.out.println("E009: Error in footer in response from Server: "+fromServer);
                 return;
              }
              everyPlayerMelodyWindow.setParticipantInput("-----------------", j++);
              everyPlayerMelodyWindow.setParticipantInput("All:", j++);
              everyPlayerMelodyWindow.setParticipantInput(netDanceListener.getCompoundMelodyString(), j++);
              everyPlayerMelodyWindow.setParticipantInput("-----------------", j++);
              everyPlayerMelodyWindow.setParticipantInput("MainPitch: "+MainPitch, j++);
              everyPlayerMelodyWindow.setParticipantInput("Duration: "+Duration, j++);
              everyPlayerMelodyWindow.setParticipantInput("Length: "+Length, j++);
              everyPlayerMelodyWindow.setParticipantInput("DecayRate: "+DecayRate, j++);
              everyPlayerMelodyWindow.setParticipantInput("ScaleType: "+ScaleType, j++);
              everyPlayerMelodyWindow.setParticipantInput("PolyphonyType: "+PolyphonyTypeStr, j++);
              everyPlayerMelodyWindow.setParticipantInput("-----------------", j++);
              NewState = true;
           }
           else
              CompoundMelody = Melody;
        }

        public void PutStateToServer(String message) throws IOException
        {
            out.println("put"+System.lineSeparator()+message);
            String fromServer = in.readLine();
            if(!fromServer.equals("Ok."))
               System.out.println("E010: Unexpected response from Server: " + fromServer);
        }

        public void put(String request, String content)
        {
           if(NetworkEstablished && qReq.size()<20)
              qReq.add(new StringPair(request, content));
        }

        public void run() 
        {
           while(alive && NetworkEstablished)
           {
              try
              {
                 StringPair request = qReq.remove();
                 if(request.name.equals("get"))
                    GetStateFromServer();
                 else
                 if(request.name.equals("put"))
                    PutStateToServer(request.content);
              }
              catch(NoSuchElementException nsee)
              {
              //do nothing, the list is empty at the moment and that's ok
              }
              catch(IOException e)
              {
                 System.out.println("E012: Communication with Server failed.");
                 NetworkEstablished = false;
                 Err = true;
                 kill();
                 System.out.println("Press 0 to exit");
                 return;
              }
           }
        }
    }

    private class ColorShowTimerTask extends TimerTask
    {
        @Override
        public void run() 
        {
             if(timecounter%101==3)
             {
                tabX=1+timecounter*41%5; tabY=1+timecounter*51%4;
             }
             if(CompoundMelody==null)
                System.out.println("E013: Compound melody not initialized");
             Graphics g = getGraphics();
             if(g==null)
                return;
             int MelodyColorIndex = timecounter%ColorScale1.length;
             Color c1 = ColorScale1[MelodyColorIndex];
             Color c2 = ColorScale2[MelodyColorIndex];
 
             int t=timecounter*91;
             for(int i=0;i<tabX;i++)
             for(int j=0;j<tabY;j++)
             {
                if(t%2==1)
                   g.setColor(c1);
                else
                   g.setColor(c2);
                g.fillRect(getWidth()/tabX*i, getHeight()/tabY*j, getWidth()/tabX, getHeight()/tabY);
                t/=2;
             }
             timecounter++;
        }
    }

    void setShowSpeed(int period)
    {
        if(colorShowTimerTask!=null)
           colorShowTimerTask.cancel();
        colorShowTimerTask = new ColorShowTimerTask();
        if(colorChangeTimer==null)
           colorChangeTimer = new Timer();
        colorChangeTimer.scheduleAtFixedRate(colorShowTimerTask, period, period);        
    }

    void updateShow(int MainPitch, int Duration, int Length, int DecayRate,
                    int ScaleType, int[] Melody )
    {
       /*
          Ideas:
          Polyphony type should play a role in the visual part.
          Show Speed should depend on something else and not just duration
             and might change independently.
       */
       ColorScale1 = new Color[Scale.length];
       ColorScale2 = new Color[Scale.length];
       int llimitR = next()/331%245;
       int dlimitR = 5+next()/331%(250-llimitR);
       int llimitG = next()/331%245;
       int dlimitG = 5+next()/331%(250-llimitG);
       int llimitB = next()/331%245;
       int dlimitB = 5+next()/331%(250-llimitB);
       for(int i=0;i<Scale.length;i++)
       {
          ColorScale1[i] = new Color(llimitR+2034*i%dlimitR, llimitG+5033*i%dlimitG, llimitB+3033*i%dlimitB, 255);
          ColorScale2[i] = new Color(255-ColorScale1[i].getRed(), 255-ColorScale1[i].getGreen(), 255-ColorScale1[i].getBlue(), 255);
       }
       setShowSpeed(Duration);
    }

    private Timer colorChangeTimer;
    ColorShowTimerTask colorShowTimerTask;
    NetDanceListener netDanceListener;
    Thread receiver;
    private int timecounter;
    private int tabX, tabY;
    int period;
    private Color[] ColorScale1, ColorScale2;
    private JFrame myFrame;

    private void setFrame(JFrame f) { myFrame = f; }

    public void ColorShowInit(int timeCounterStart, int period) 
    {
        timecounter = timeCounterStart;
        tabX=1+timecounter*41%5; 
        tabY=1+timecounter*51%4;
        colorChangeTimer = null;
        colorShowTimerTask = null;
        updateShow((int)MainPitch, Duration, Length, DecayRate, ScaleType, CompoundMelody);
    }

    public void frameInit() 
    {
        int timeCounterStart = 1;
        JFrame frame = new JFrame("Simple Sketching Program");
        frame.getContentPane().add(this, BorderLayout.CENTER);
        frame.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
        frame.setExtendedState(JFrame.NORMAL); 
         frame.setSize(800, 500);
        frame.setUndecorated(false);
        frame.setVisible(true);
        setFrame(frame);
    }

/* ------SOUND----------------------------------------------------------- */

   public static final int CHOIRCOUNT = 1;
   public static final int SCALE6LEN = 15;
   public static final int SCALE10LEN = 20;

/* This window is shown only if the performance is networked and it shows the compound melody
   and the individual melodies */

   class MagickSquareWindow extends JPanel
   {
      private JFrame myFrame;
      private JTextArea compoundMelody;

      public void setCompoundMelody(String cm) 
      {
         int i, maxDigitCnt=0, rowlength; 
         String[] numbers = cm.split("\\s+");
         for(i=0;i<numbers.length;i++)
            if(numbers[i].length()>maxDigitCnt)
               maxDigitCnt = numbers[i].length();
         rowlength = (int)Math.floor(Math.sqrt(numbers.length));
         StringBuilder sb = new StringBuilder();
         for(i=0;i<numbers.length;i++)
         {
            if(i!=0&&i%rowlength==0)
               sb.append(System.lineSeparator());
            sb.append(String.format("%1$"+maxDigitCnt+ "s ", numbers[i]));
         }
         compoundMelody.setText(sb.toString()); 
      }

      public MagickSquareWindow() 
      {
         JFrame frame = new JFrame("Conjunction");
         compoundMelody = new JTextArea();
         Font labelFont = compoundMelody.getFont();
         Font frameFont = new Font(labelFont.getName(), Font.PLAIN, 20);
         compoundMelody.setFont(frameFont);
         frame.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
         frame.setExtendedState(JFrame.NORMAL); 
         frame.setUndecorated(false);
         frame.setSize(500, 500);
         frame.getContentPane().add(compoundMelody,BorderLayout.CENTER);
         frame.setVisible(true);
         frame.toFront();
         myFrame = frame;
      }
   }

   class CompoundMelodyWindow extends JPanel
   {
      private JFrame myFrame;
      private JLabel compoundMelody;
      private JButton btnSave;
      private JTextArea commandArea;
      private JPanel buttonRow;
      private JMenuBar menuBar;
      JMenu menu;
      JMenuItem menuItemSave;

      public void setCompoundMelody(String cm) { compoundMelody.setText(cm); }

      private int getParamsEnd(String input)
      {
         int paramsEnd = input.lastIndexOf('/');
         if(paramsEnd>=0)
         {
            do
            { paramsEnd++; }
            while(input.charAt(paramsEnd)>='0'&&input.charAt(paramsEnd)<='9');
         }
         return paramsEnd;
      }

      public void updateCommandArea(String input)
      {
         //If the user selected some text then do not update the field, because
         //selected area will be lost
         if( commandArea.getSelectedText() != null )
            return;         
         int caretPos = commandArea.getCaretPosition();
         int paramsEndNew = getParamsEnd(input);
         //otherwise update only the area of common params
         String oldText = commandArea.getText();
         //if the cursor is in the area of common params, then do not update anything
         //Of course if two players try updating the common params at the same time,
         //the second will overwrite the first, even if they update two different common params
         if(oldText.length()>0&&caretPos<=paramsEndNew)
            return;
         int paramsEndOld = getParamsEnd(oldText);
         if(paramsEndNew<0)
         {
            System.out.println("E014: No fractions in Server input: "+input);
            return;
         }
         if(paramsEndOld<0)
            commandArea.setText(input);
         else
            commandArea.setText(input.substring(0, paramsEndNew) + oldText.substring(paramsEndOld));
         try
         {
            commandArea.setCaretPosition(caretPos);
         }
         catch(IllegalArgumentException e)
         {
            commandArea.setCaretPosition(commandArea.getDocument().getLength());
         }
      }

      class ApplyActionListener implements ActionListener
      {
      public void actionPerformed(ActionEvent e)
      {
         try
         {
            FileWriter fw = new FileWriter("input"+index+".txt");
            BufferedWriter bw = new BufferedWriter(fw);
            bw.write(commandArea.getText());
            bw.close();
            fw.close();
         }
         catch(IOException ioe)
         {
            System.out.println("E015: Cannot save to input" + index + ".txt");
         }
      }
      }

      public CompoundMelodyWindow() 
      {
         JFrame frame = new JFrame("Compound Melody");
         compoundMelody = new JLabel("0 0 0 0", SwingConstants.CENTER);
         commandArea = new JTextArea();
         Font labelFont = compoundMelody.getFont();
         Font frameFont = new Font(labelFont.getName(), Font.PLAIN, 20);
         compoundMelody.setFont(frameFont);
         commandArea.setFont(frameFont);
         btnSave = new JButton("Save");
         btnSave.setMnemonic(KeyEvent.VK_S);
         btnSave.addActionListener(new ApplyActionListener());
         menuBar = new JMenuBar();
         menu = new JMenu("File");
         menu.setMnemonic(KeyEvent.VK_F);
         menuBar.add (menu);
         menuItemSave = new JMenuItem("Save", KeyEvent.VK_V);
         menuItemSave.setAccelerator(KeyStroke.getKeyStroke(
            KeyEvent.VK_S, Toolkit.getDefaultToolkit().getMenuShortcutKeyMask()));
         menuItemSave.addActionListener(new ApplyActionListener());
         menu.add(menuItemSave);
      
         buttonRow = new JPanel();
         buttonRow.setLayout(new GridLayout(1, 6));
         buttonRow.add(btnSave);
         frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
         frame.setExtendedState(JFrame.NORMAL); 
         frame.setUndecorated(false);
         frame.setSize(500, 500);
         frame.setLocation(0,0);

         frame.getContentPane().add(compoundMelody,BorderLayout.NORTH);
         frame.getContentPane().add(commandArea, BorderLayout.CENTER);
         frame.getContentPane().add(buttonRow, BorderLayout.SOUTH);
         frame.setJMenuBar(menuBar);
         frame.setVisible(true);
         frame.toFront();
         myFrame = frame;
      }
   }

   class EveryPlayerMelodyWindow extends JPanel
   {
      private JFrame myFrame;
      private JList inMelodies;
      DefaultListModel<String> participantInput;
      String participantNickname;

      public void setParticipantNickname(String name)
      {
         participantNickname = name;
      }

      public int getLineCnt()
      {
         return participantInput.size();
      }

      public String getName() { return participantNickname; }
 
      public void toggleVisible() 
      {
         myFrame.setVisible(!myFrame.isVisible());
         if(myFrame.isVisible())
             myFrame.toFront(); 
      }

      public void setParticipantInput(String inputline, int pos)
      {         
         if(pos<participantInput.getSize()&&pos>=0)
            participantInput.setElementAt(inputline, pos);
         else
            participantInput.addElement(inputline);
      }

      public void close() { myFrame.setVisible(false); }

      public EveryPlayerMelodyWindow()
      {
         participantNickname = "All players";
         JFrame frame = new JFrame(participantNickname + " input");
         participantInput = new DefaultListModel<String>();
         inMelodies = new JList<String>(participantInput);
         inMelodies.setFont(new Font("Courier New", Font.PLAIN, 20));
         frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
         frame.setExtendedState(JFrame.NORMAL); 
         frame.setUndecorated(false);
         frame.setSize(500,500);
         frame.setLocation(500, 0);
         frame.getContentPane().add(inMelodies, BorderLayout.CENTER);
         frame.setVisible(true);

         myFrame = frame;
      }
   }

   MagickSquareWindow magickSquareWindow;
   CompoundMelodyWindow compoundMelodyWindow;
   EveryPlayerMelodyWindow everyPlayerMelodyWindow;

//This Socket will be used to communicate to the server
   Socket kkSocket = null;

//This Socket will be used to communicate to the download server on another client
   Socket kkSocketForDownload = null;

//This is something we use to write to the socket
   PrintWriter out = null;

//This is something we use to read from the socket
   BufferedReader in = null;

//a boolean variable used in some loops
   boolean unsuccessful = true;

//a structure used to discover name of this client on the Internet
   InetAddress myComputerNameAndIP = null;

//name or IP address of the server (configurable in file ClientConfiguration.txt)
   String serverName = null;

//Port number (configurable in file ClientConfiguration.txt)
   int PortNumber = 4444;

//nickname (configurable in file ClientConfiguration.txt)
   String nickName = null;

   boolean NetworkEstablished = false;
   int Identifier = 0;

   AudioFormat format;
   DataLine.Info sourceInfo;
   SourceDataLine sourceLine;
   byte[] targetData;
   int bufferLen;
   boolean alive;
   boolean Muted, Verbose, Show, Log, Magick;
   int SoundMax;
   int Iteration = 0;
   String logDirName;
   FileOutputStream fosRec;
   PrintWriter pwLog;   
   String outFileRec, outFileLog;
   int index;
   double MainPitch;
   double [] Scale;
   int Duration, Length, DecayRate, ScaleType, Span;
   int[] PolyphonyType;
   //The current melody of the this performer
   int[] Melody;
   //The current combined melody of all performers - this is returned from the Server
   int[] CompoundMelody;
   boolean hasEnded;
   boolean Err;

   public boolean HasEnded() { return hasEnded; }
   public boolean HasError() { return Err; }

   public static final double[] Scale6={
      1.0/1, 6.0/5, 5.0/4, 4.0/3, 3.0/2, 
      8.0/5, 5.0/3, 2.0/1, 12.0/5, 10.0/4, 
      8.0/3, 6.0/2, 16.0/5, 10.0/3, 4.0/1    
   };
   
   public static final double[] Scale10={
      1.0/1, 10.0/9, 9.0/8, 8.0/7, 7.0/6, 
      6.0/5, 5.0/4, 9.0/7, 4.0/3, 7.0/5, 
      10.0/7, 3.0/2, 14.0/9, 8.0/5, 5.0/3, 
      12.0/7, 7.0/4, 16.0/9, 9.0/5, 2.0/1 
   };
   
   public void EstablishNetworkConnection() throws IOException
   {
//The client firstly reads the ClientConfiguration.txt file
        try
        {
           FileReader  ul = new FileReader ("ClientConfiguration.txt");
           BufferedReader inFile = new BufferedReader(ul);
           String item;
           try
           {
//ClientConfiguration.txt file contains the name of the name server
//and the user's nickname
              String serverNameAndPort = inFile.readLine ();
              String [] serverNameAndPortParts = serverNameAndPort.split(":");
              if(serverNameAndPortParts.length!=2)
              {
                 System.out.println("E016a: Server name or IP or port not specified, starting in standalone mode.\n");
                 return;
              }
              serverName = serverNameAndPortParts[0];
              System.out.println( "Server IP = "+serverName );
              PortNumber = Integer.parseInt(serverNameAndPortParts[1]);
              nickName = inFile.readLine ();
              System.out.println( "Nick = "+nickName );
              while(true)
              {
                 String OptionalLine = inFile.readLine();
                 if(OptionalLine == null)
                    break;
                 if(OptionalLine.startsWith("Mute"))
                    Muted = true;
                 else
                 if(OptionalLine.startsWith("Verbose"))
                    Verbose = true;
                 else
                 if(OptionalLine.startsWith("Show"))
                    Show = true;
                 else
                 if(OptionalLine.startsWith("Log"))
                    Log = true;
                 else
                 if(OptionalLine.startsWith("Magick"))
                    Magick = true;
                 if(OptionalLine.startsWith("SoundMax"))
                 {
                    SoundMax = 32767;
                    String[] line = OptionalLine.split("\\s+");
                    if(line.length!=2)
                       System.out.println("E016b: Error in line: SoundMax n");
                    else
                       SoundMax = Integer.parseInt(line[1]);
                 }
              }
              inFile.close();
              ul.close ();
              if(serverName.startsWith("-"))
              {
                 System.out.println("E016c: Server name or IP not specified, starting in standalone mode.\n");
                 return;
              }
           }
           catch(NumberFormatException nfe)
           {
              System.out.println("E017a: Invalid Port");
              return;
           }
           catch (IOException g)
           {
              System.out.println("E017b: Invalid ClientConfiguration.txt");
              return;
           }
        }
        catch (IOException g)
        {
           System.out.println("E018: Couldn't read from file ClientConfiguration.txt");
           return;
        }
//Afterwards, the client tries to discover his own name on the Internet
        try
        {
           myComputerNameAndIP = InetAddress.getLocalHost();
           System.out.println( "My computer's name and IP are " + myComputerNameAndIP );
        }
//If he fails, reports an error and exits.
        catch (UnknownHostException e)
        {
           System.err.println("Don't know my name, sorry");
           return;
        }

//Then opens a client connection to the name server
        try {
            kkSocket = new Socket( serverName , PortNumber);
            out = new PrintWriter(kkSocket.getOutputStream(), true);
            in = new BufferedReader(new InputStreamReader(kkSocket.getInputStream()));
        } catch (UnknownHostException e) {
            System.err.println("Don't know about host: "+serverName );
            return;
        } catch (IOException e) {
            System.err.println("Couldn't get I/O for the connection to: " +serverName);
            return;
        }
//The client sends to the server his identification, in the format nick:name:port
        out.println( nickName );

        String assignedId = "0";
        try
        {
           assignedId = in.readLine();
           Identifier = Integer.parseInt(assignedId);
        }
        catch( NumberFormatException e)
        {
           System.out.println("E019: Could not parse the assigned identifier in response from Server: "+assignedId);
           return;
        }

//The client receives the welcome message from the server
        System.out.println("Server: Your identifier is " + Identifier);
        System.out.println("Server: " + in.readLine());

       NetworkEstablished = true;
   }

/* Closes the communication with the server on exit */
   public void CloseNetworkConnection() throws IOException
   {
        out.println("bye");
        System.out.println("Server: " + in.readLine());
        out.close();
        in.close();
        kkSocket.close();
   }
   
   public OneBitNetDance(int i)
   {
      index = i;
      a=123004463; 
      b=715311001; 
      c=223094001;
      bufferLen=196400;
      format = new AudioFormat(44100, 16, 1, true, true);
      sourceInfo = new DataLine.Info(SourceDataLine.class, format);
      targetData = new byte[bufferLen];
      alive = true;
      MainPitch = 300.0;
      Duration = 58;
      Length = 200;
      DecayRate = 4;
      ScaleType = 6;
      Scale = Scale6;
      Span = 4;
      colorShowTimerTask = null;
      hasEnded = false;
      Err = false;
      Muted = false;
      Verbose = false;
      Show = false;
      Log = false;
      Magick = false;
      SoundMax = 32767;
      Melody = new int[4];
      for(int j=0;j<4;j++)
         Melody[j] = 0;
      PolyphonyType = new int[] {1, 1, 1, 1, 2, 1};
      
      CompoundMelody = Melody;
      try
      {
         EstablishNetworkConnection();
      }
      catch(IOException e)
      {
         System.out.println("E020: Cannot find server.");
      }
      if(NetworkEstablished)
      {
         compoundMelodyWindow = new CompoundMelodyWindow();
         everyPlayerMelodyWindow = new EveryPlayerMelodyWindow();
         if(Magick)
            magickSquareWindow = new MagickSquareWindow();
      }
      netDanceListener = new NetDanceListener();
      receiver = new Thread(netDanceListener, "0");
      receiver.start();
      
      if(Log)
         initLog();
      readState();
      if(Show)
      {
         frameInit();
         ColorShowInit(Length, Duration);
      }
   }

   boolean modified;
   private long timeStamp;

   private String showTime()
   {
       Calendar cal = Calendar.getInstance();
       cal.getTime();
       SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd-HH-mm-ss");
       String currTime = sdf.format(cal.getTime());
       System.out.println( currTime );
       return currTime;
   }

   private void initLog()
   {
      logDirName = showTime();
      File logDir = new File(logDirName);
      if(!logDir.exists()) 
      {
         try
         {
            logDir.mkdir();
         } 
         catch(SecurityException se)
         {
            System.out.println("E021: Cannot create directory " + logDirName);
            return;
         }        
      }
      outFileRec = logDirName + "/" + index;
      outFileLog = logDirName + "/" + index + ".txt";
      try
      {
         fosRec = new FileOutputStream(outFileRec);
         pwLog = new PrintWriter(outFileLog, "ASCII");
      }
      catch(IOException e)
      {
         System.out.println("E022: Could not open file " + outFileRec + " or " + outFileLog);
      }
   }

   private int NumerologicalSum(int n)
   {
      int sum=0;
      while(n>0)
      {
         sum+=n%10;
         n-=n%10;
         n/=10;
      }
      return sum;
   }

   private int NumerologicalReduction(int n, int threshold)
   {
      while(n>threshold)
         n=NumerologicalSum(n);
      return n;
   }

   public void updateState(int[] inNum, boolean fractions, String completeInput)
   {
      int i;
      MainPitch = (double)inNum[0];
      Duration = inNum[1];
      Length = inNum[2];
      DecayRate = inNum[3];
      ScaleType = 6; //inNum[4];
      if(fractions)
      {
         i=5;
         while(inNum[i]!=0)
            i++;
         PolyphonyType = Arrays.copyOfRange(inNum, 5, i);
         i++;
      }
      else
      {
         System.out.println("W001: Warning: this is deprecated and should not be used.");
         switch(inNum[5])
         {
         case 0: PolyphonyType = new int[]{1, 1, 1, 1, 1, -1}; break;
         case 1: PolyphonyType = new int[]{1, 1, 1, 1, 2, 1}; break;
         case 2: PolyphonyType = new int[]{1, 1, 1, 2, 3, 1}; break;
         default:
            System.out.println("E023: Invalid polyphony type " + inNum[5] + ", set to 1");
            PolyphonyType = new int[]{1, 1, 1, 1, 2, 1};
         }
         i=6;
      }
      Melody = Arrays.copyOfRange( inNum, i, inNum.length);
      if(!NetworkEstablished)
         CompoundMelody = Melody;
      if(ScaleType==10)
         Scale = Scale10;
      else
         Scale = Scale6;

      if(NetworkEstablished)
      {
//Sending the six parameters to the server
         String params = inNum[0]+" "+inNum[1]+" "+inNum[2]+" "+inNum[3]+" "+inNum[4]+" ";
         StringBuilder sbParams = new StringBuilder(params);
         for(i=0;i<PolyphonyType.length/3;i++)
            sbParams.append(PolyphonyType[3*i+2]==-1?"-":""
               +PolyphonyType[3*i]+"/"+PolyphonyType[3*i+1]
               +(i==PolyphonyType.length/3-1?"":" "));
//Sending the melody without comments to the server
         StringBuilder sbMelody = new StringBuilder();
         for(i=0;i<Melody.length;i++)
            sbMelody.append(Melody[i]+(i==Melody.length-1?"":" "));
         String message = 
            sbParams.toString()+System.lineSeparator()
            +sbMelody.toString()+System.lineSeparator()
            +completeInput;
         if(!message.endsWith(System.lineSeparator()))
            message+=System.lineSeparator();
         message+="end lines";
         netDanceListener.put("put", message);
      }

      updateShow((int)MainPitch, Duration, Length, DecayRate, ScaleType, CompoundMelody);
   }

   void writeLog()
   {
      int i;
      if(pwLog==null||Melody==null)
         return;
      pwLog.print(
         MainPitch + " " +
         Duration + " " +
         Length + " " +
         DecayRate + " " +
         ScaleType + " "
      );
      for(i=0;i<PolyphonyType.length;i+=3)
         pwLog.print((PolyphonyType[i+2]==-1?"-":"") + PolyphonyType[i] + "/" + PolyphonyType[i+1] + " ");
      for(i=0;i<Melody.length;i++)
         if(Melody[i]<0)
            pwLog.print("_ ");
         else
            pwLog.print(Melody[i] + " ");
      pwLog.println("-1");
   }

   private boolean readState()
   {
      File file;
      long newTimeStamp;
      boolean result = false;
      BufferedReader br = null;
      boolean fractions = false;

//First check if the file has been modified
      file = new File("input"+index+".txt");
      newTimeStamp = file.lastModified();
      if(newTimeStamp!=timeStamp)
      {
          result = true;
          timeStamp = newTimeStamp;
          if(Log) writeLog();
      }
      else
//If it hasn't been modified, do nothing.
         return false;
      try 
      {
         int state = 0;
//The file has been modified. Read all of it into one long string.
         FileReader fr = new FileReader("input0.txt");
         br = new BufferedReader(fr);
         StringBuilder sb = new StringBuilder();
         String line = br.readLine();

         while (line != null) 
         {
            sb.append(line);
            sb.append(System.lineSeparator());
            line = br.readLine();
         }
         String completeInput = sb.toString();
         br.close();

//Split by anything non-alphanumeric, excluding / and -, because we can have 
//fractions and negative numbers
         String[] numstrs=completeInput.split("[\\W+&&[^/]&&[^-]]");
//Input has to have at least seven elements: 6 common parameters including at least one 
//fraction for polyphony type and at least one tone for the melody
         if(numstrs.length<7)
         {
            System.out.println("E011: Input corrupted: " + completeInput);
            return false;
         }
         int finalCount = 0;
//This function is going to produce an array of integers based on the above file.
//Each fraction will be three integers: dividend, divisor and sign.
//We first need to establish how many numbers are there in the file, including negative
//numbers, fractions and _. Fractions count three times.
//Plus, in the function that receives the list of integers, we need a signal that
//the fractions stop at a certain point and normal numbers proceed (the melody).
//Thus we put a zero at the end of the list of fractions. Se each fraction counts three
//times, plus one zero after the entire list of fractions. We don't need to begin
//the list of fractions with a similar marker because we know it starts from sixth
//number onwards.
         for(int i = 0;i<numstrs.length;i++)
            if(numstrs[i].length()>0&&(numstrs[i].charAt(0)=='_'||numstrs[i].charAt(0)=='-'||Character.isDigit(numstrs[i].charAt(0))))
            {
               finalCount++;
               if(numstrs[i].indexOf('/')!=-1)
               {
                  finalCount+=2;
                  state = 1;
               }
               else
               {
                  if(state==1)
                     finalCount++;
                  state = 0;
               }
            }
//Create the array of integers of the cardinality established above
         int[] numbers = new int[finalCount];
         int j=0;
         state = 0;
         for(int i = 0;i<numstrs.length;i++)
         {
            if(numstrs[i].length()>0)
            {
//Underscores (_) are represented by a -1
//If state is 1, and we see a _, then the fractions section is over and we have to put a 0 marker at the end of it
               if(numstrs[i].charAt(0)=='_')
               {
                  if(state==1)
                  {
                     numbers[j++]=0;
                     state = 2;
                  }
                  numbers[j++]=-1;
               }
               else
//If it's a number
               if(Character.isDigit(numstrs[i].charAt(0))||numstrs[i].charAt(0)=='-')
               {
//If it's a fraction: fractions can only appear in the section Polyphony Type and nowhere else
                  if(numstrs[i].indexOf('/')!=-1)
                  {
//state==2 means post Polyphony Type; j<5 means Polyphony Type hasn't been reached yet.
//If there is an error, report it. If not break the fraction into three integers:
//dividend, divisor and sign.
                     if(state==2||j<5)
                     {
                        System.out.println("E024: Unexpected fraction " + numstrs[i] + " at position " + j + ", ignored.");
                     }
                     else
                     {
                        String[] components = numstrs[i].split("/");
                        try
                        {
                           numbers[j] = NumerologicalReduction(Math.abs(Integer.parseInt(components[0])), 9);
                           if(numbers[j]==0)
                              numbers[j]=1;
                           j++;
                           numbers[j] = NumerologicalReduction(Integer.parseInt(components[1]), 9);
                           if(numbers[j]==0)
                              numbers[j]=1;
                           j++;
                        }
                        catch(NumberFormatException nfe)
                        {
                           System.out.println("E029: Invalid fraction "+numstrs[i]);
                           return false;
                        }
                        numbers[j++] = (numstrs[i].charAt(0)=='-'?-1:1);
                        fractions = true;
                        state = 1;
                     }
                  }
                  else
                  {
//If it's not a fraction
                     if(state==1)
                     {
//If it's a number after the list of fractions, then update state to 2 and insert
//a zero to end the list of fractions before the current number.
                        numbers[j++]=0;
                        state = 2;
                     }
                     else
                     {
                        if(j==5)
                        {
//This is a deprecated style of stating the polyphony type. However it can be used unless it's a networked
//performance, where the parameters are shared in common and fractions are essential to be able to tell
//where the common params section ends
                           if(NetworkEstablished)
                           {
                              System.out.println("E025: Expected at least one fraction at position " + j + ", found "+numstrs[i]+".");
                              return false;
                           }
                        }
                     }
//place the number in its place
                     try
                     {
                        numbers[j]=Integer.parseInt(numstrs[i]);
                     }
                     catch(NumberFormatException nfe)
                     {
                        System.out.println("E030: Invalid number "+numstrs[i]);
                        return false;
                     }
                     if(j<5&&j!=2)
                        numbers[j]=NumerologicalReduction(numbers[j], 1000);
                     if(j<5&&numbers[j]==0)
                        numbers[j]=0;
//This is a melody number so it can't be negative.
                     if(numbers[j]<0)
                     {
                        System.out.println("E026: Unexpected negative number "+numbers[j]+", changed sign.\n");
                        numbers[j]=-numbers[j];
                     }
                     j++;
                  }
               }
            }
         }

//Update the state of the system with the parameters read from the file.
         if(state!=2)
         {
            System.out.println("E031: Input corrupted: missing melody.");
            return false;
         }
         updateState(numbers, fractions, completeInput);
         modified=result;
      }
      catch (IOException e) {
         e.printStackTrace();
      } 
      return result;
   }

   void oneBitify()
   {
      for(int i=0;i<bufferLen;i+=2)
      {
         short sample = (short)(targetData[i]*(byte)256+targetData[i+1]);
         sample = (short)(sample>0?SoundMax:-(SoundMax+1));
         targetData[i]=(byte)(sample>>8);
         targetData[i+1]=(byte)(sample&255);
      }
   }

   void saveBuffer()
   {
      try
      {
         fosRec.write(targetData);
      }
      catch(IOException e)
      {
         System.out.println("E027: Could not write to file " + outFileRec);
      }
   }

   void placeSample(int pos, int value)
   {
      int existingVal = targetData[pos]*256+targetData[pos+1];
      existingVal+=value;
      targetData[pos]=(byte)(existingVal>>8);
      targetData[pos+1]=(byte)(existingVal&255);
   }

   public static int getSampleCountFromMs(int ms)
   {
      return (int)((double)ms*44.1);
   }

   public static int getSampleCountFromHz(double hz)
   {
      return (int)(44100.0/hz);
   }


   void placeNote(int pos_ms, int d_ms, int note, double mainPitch, int sign)
   {
      if(note<0)
         return;
      int pos = 2*getSampleCountFromMs(pos_ms);
      int duration = getSampleCountFromMs(d_ms);
      double frequency = mainPitch*(sign==1?Scale[note]:Scale[Scale.length-1-note]/Span);
      int wavelen = getSampleCountFromHz(frequency);
      //If frequency is too high, no sound is heard. If this distinction is not made here
      //then the below loop never exits.
      if(wavelen==0)
         return;
      int i=pos, k, l=0;
      double j=0.0;
      double decay = (double)wavelen*wavelen/DecayRate/duration;
      while(l<duration)
      {
         for(k=0;k<wavelen/2+j&&l<duration;k++)
         {
            if(i>=bufferLen)
               i-=bufferLen;
            placeSample(i, 1000);
            i+=2; l++;
         }
         for(k=0;k<wavelen/2-j&&l<duration;k++)
         {
            if(i>=bufferLen)
               i-=bufferLen;
            placeSample(i, -1000);
            i+=2; l++;
         }
         j+=decay;
      }
   }

   private void getResultsFromServer()
   {
      int i, j;
      if(!NetworkEstablished)
      {
         return;
      }
      if(!netDanceListener.isNewState())
         return;
      String[] names = netDanceListener.getNameList();
      int[] indexes = netDanceListener.getIndexList();
      if(names==null)
         return;
      MainPitch = netDanceListener.getMainPitch();
      Duration = netDanceListener.getDuration();
      Length = netDanceListener.getLength();
      DecayRate = netDanceListener.getDecayRate();
      ScaleType = netDanceListener.getScaleType();
      PolyphonyType = netDanceListener.getPolyphonyType();
      CompoundMelody = netDanceListener.getCompoundMelody();
      compoundMelodyWindow.setCompoundMelody(netDanceListener.getCompoundMelodyString());
      if(Magick)
         magickSquareWindow.setCompoundMelody(netDanceListener.getCompoundMelodyString());
      netDanceListener.UpdateState();
   }

   void fillBuffer()
   {
      int skipLen;
      if( CompoundMelody.length<8 )
         skipLen = 8/CompoundMelody.length + 1;
      else
         skipLen = 1;
      if(Iteration%skipLen==0)
      {
         netDanceListener.put("get", "");
      }
      getResultsFromServer();
      bufferLen = PolyphonyType.length/3*getSampleCountFromMs(2*Duration)*CompoundMelody.length;
      targetData = new byte[bufferLen];
      Arrays.fill(targetData, (byte)0);
      if(Muted)
      {
         for(int i=0;i<bufferLen;i+=2)
         {
            targetData[i]=(byte)(0);
            targetData[i+1]=(byte)(0);
         }
      }
      else
      {
         for(int i=0;i<CompoundMelody.length;i++)
         {
//Change for Malgorzata
            CompoundMelody[i] = NumerologicalReduction(CompoundMelody[i], Scale.length-1);
            for(int j=0;j<PolyphonyType.length;j+=3)
            {
               int P=PolyphonyType.length/3; int[] PT = PolyphonyType;
               placeNote(Duration*(P*i+j/3), Length, CompoundMelody[i], MainPitch*PT[j]/PT[j+1], PT[j+2]); 
            }
         }
         oneBitify();
      }
      Iteration++;
   }

/*
   The order of calling the server is:
   1) put
   2) get
*/

   public void run()
   {
      try
      {
         sourceLine = (SourceDataLine) AudioSystem.getLine(sourceInfo);
         sourceLine.open(format);
         sourceLine.start();
         while(alive)
         {
            fillBuffer();
            sourceLine.write(targetData, 0, bufferLen);
            if(Log) saveBuffer();
            readState();
         }
         sourceLine.write(targetData, 0, bufferLen);
         sourceLine.stop();
         sourceLine.close();
         if(Log)
         {
            saveBuffer();
            fosRec.close();
            writeLog();
            pwLog.close();
         }
         if(Show)
         {
            colorShowTimerTask.cancel();
            colorChangeTimer = null;
            myFrame.setVisible(false);
            myFrame.dispatchEvent(new WindowEvent(myFrame, WindowEvent.WINDOW_CLOSING));
         }
         
      }
      catch (Exception e) 
      {
         e.printStackTrace();
      }
      hasEnded = true;
   }

   void kill()
   {
      System.out.println("Stopping program...");
      bufferLen = 2*bufferLen;
      targetData = new byte[bufferLen];
      Arrays.fill(targetData, (byte)0);
      fillBuffer();
      alive = false;
      netDanceListener.kill();
      try
      {
         if(NetworkEstablished)
            CloseNetworkConnection();
      }
      catch(IOException e)
      {
         System.out.println("E028: Could not close connection to the server.");
      }
   }

   public static void main(String[] args) 
   {
      int i;

      System.out.println("Press 0 to exit");
      OneBitNetDance dance = new OneBitNetDance(0);
      Thread music = new Thread(dance, "0");
      music.start();

      String x;

      Scanner s = new Scanner(System.in);

      while (true) 
      {
         x = s.nextLine();
         if(x.length()>0 && x.charAt(0)=='0')
            break;
      }
      if(!dance.HasError())
         dance.kill();
      while(!dance.HasEnded())
      {
         try
         {
            Thread.sleep(100);
         }
         catch (InterruptedException e)
         {
            System.out.println("Main thread interrupted.");
         }
      }
      System.exit(0);
   }
}

